package org.khelekore.jsizer;

import java.io.DataInputStream;
import java.io.IOException;
import java.net.URLClassLoader;
import java.util.HashSet;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.commons.EmptyVisitor;
import org.objectweb.asm.Type;

/** A ClassSizer that uses the asm library to size classes.
 */
public class AsmSizer implements ClassSizer {

    private byte[] buf = new byte[64 * 1024];

    public void setup (URLClassLoader cl) {
	// Nothing to do.
    }

    public ClassSize size (JarFile jf, JarEntry je, String cname, JSizer sizer) {
	DataInputStream is = null;
	int size = -1;
	try {
	    is = new DataInputStream (jf.getInputStream (je));
	    try {
		size = (int)je.getSize ();
		if (buf.length < size)
	    buf = new byte[size];
		is.readFully (buf, 0, size);
	    } finally {
		is.close ();
	    }
	    ClassSizeVisitor sv = new ClassSizeVisitor ();
	    ClassReader cr = new ClassReader (buf, 0, size);
	    cr.accept (sv, 0);
	    return new ClassSize (cname, sv.getSuperName (),
				  sv.getAccess (),
				  sv.getNumFields (),
				  sv.getNumMethods (),
				  sv.getNumMethodLines (),
				  sv.usedClasses.size ());
	} catch (IOException e) {
	    System.err.println ("failed to read: " + je + ", size: " + size);
	    e.printStackTrace ();
	}
	return null;
    }

    private static class ClassSizeVisitor extends EmptyVisitor {
	private String superName;
	private int access;
	private int numFields;
	private int numMethods;
	private int numMethodLines;
	private Set<String> usedClasses = new HashSet<String> ();

	@Override public void visit(int version, int access, String name,
				    String signature, String superName,
				    String[] interfaces) {
	    this.superName = superName;
	    this.access = access;
	}

	public String getSuperName () {
	    return superName;
	}

	public int getAccess () {
	    return access;
	}

	public int getNumFields () {
	    return numFields;
	}

	public int getNumMethods () {
	    return numMethods;
	}

	public int getNumMethodLines () {
	    return numMethodLines;
	}

	public void addLineNumbers (int numLines) {
	    numMethodLines += numLines;
	}

	private void addUsedClass (Type t) {
	    if (t.getSort () == Type.OBJECT)
		usedClasses.add (t.getClassName ());
	}

	private void addUsedClass (String desc) {
	    if (desc == null)
		return;
	    addUsedClass (Type.getType (desc));
	}

	private void addUsedClassesByMethod (String methodDesc) {
	    Type[] types = Type.getArgumentTypes (methodDesc);
	    for (Type t : types)
		addUsedClass (t);
	}

	@Override public FieldVisitor visitField (int access,
						  String name,
						  String desc,
						  String signature,
						  Object value) {
	    addUsedClass (desc);
	    numFields++;
	    return null;
	}

	@Override public MethodVisitor visitMethod (int access, String name,
						    String desc,
						    String signature,
						    String[] exceptions) {
	    addUsedClassesByMethod (desc);
	    numMethods++;
	    MethodUsageCounter lc = new MethodUsageCounter (this);
	    return lc;
	}
    }

    private static class MethodUsageCounter extends EmptyVisitor {
	private ClassSizeVisitor csv;
	private Set<Integer> lineNumbers = new HashSet<Integer> ();

	public MethodUsageCounter (ClassSizeVisitor csv) {
	    this.csv = csv;
	}

	@Override public void visitLineNumber (int line, Label start) {
	    lineNumbers.add (line);
	}

	@Override public void visitEnd () {
	    csv.addLineNumbers (lineNumbers.size ());
	}

	@Override public void visitFieldInsn(int opcode,
					     String owner,
					     String name,
					     String desc) {
	    csv.addUsedClass (Type.getObjectType (owner));
	    csv.addUsedClass (desc);
	}

	@Override public void visitLocalVariable (String name,
						  String desc,
						  String signature,
						  Label start,
						  Label end,
						  int index) {
	    csv.addUsedClass (desc);
	}

	@Override public void visitMethodInsn (int opcode,
					       String owner,
					       String name,
					       String desc) {
	    csv.addUsedClass (Type.getObjectType (owner));
	    csv.addUsedClassesByMethod (desc);
	}

	@Override public void visitMultiANewArrayInsn (String desc,
						       int dims) {
	    csv.addUsedClass (Type.getType (desc));
	}

	@Override public void  visitTypeInsn (int opcode,
					      String type) {
	    csv.addUsedClass (Type.getObjectType (type));
	}

	@Override public void visitLdcInsn (Object cst) {
	    if (cst instanceof Type) {
		csv.addUsedClass ((Type)cst);
	    }
	}
    }
}
