package org.khelekore.jsizer;

import org.objectweb.asm.Opcodes;

/** information about a class. */
public class ClassSize {
    private final String className;
    private final String superName;
    private final int access;
    private final int numFields;
    private final int numMethods;
    private final int numMethodLines;
    private final int numUsedClasses;

    public ClassSize (String className, String superName, int access,
		      int numFields, int numMethods, int numMethodLines, 
		      int numUsedClasses) {
	if (className == null)
	    throw new IllegalArgumentException ("name may not be null");
	this.className = className;
	if (superName == null && !className.equals ("java.lang.Object"))
	    throw new IllegalArgumentException ("superName may not be null: " +
						className);
	this.superName = superName;
	this.access = access;
	if (numFields < 0)
	    throw new IllegalArgumentException ("numFields may not be " +
						"negative: " + numFields);
	this.numFields = numFields;
	if (numMethods < 0)
	    throw new IllegalArgumentException ("numMethods may not be " +
						"negative: " + numMethods);
	this.numMethods = numMethods;
	if (numMethodLines < 0)
	    throw new IllegalArgumentException ("numMethodLines may not be " +
						"negative: " + numMethodLines);
	this.numMethodLines = numMethodLines;
	this.numUsedClasses = numUsedClasses;
    }

    @Override public String toString () {
	return getClass ().getSimpleName () + "{className: " + className +
	    ", superName: " + superName + ", numFields: " + numFields +
	    ", numMethods: " + numMethods + ", numMethodLines: " +
	    numMethodLines + ", access: " + access +
	    " (" + getAccessNames () + ")}";
    }

    public String getClassName () {
	return className;
    }

    public String getSuperName () {
	return superName;
    }

    private boolean hasAccessFlag (int flag) {
	return (access & flag) == flag;
    }

    private String getAccessNames () {
	StringBuilder ret = new StringBuilder ();
	if (isPublic ())
	    ret.append ("public");

	if (isFinal ())
	    addWithBlank (ret, "final");

	if (isInterface ())
	    addWithBlank (ret, "interface");
	else if (isAbstract ())
	    addWithBlank (ret, "abstract");

	return ret.toString ();
    }

    private void addWithBlank (StringBuilder sb, String s) {
	if (sb.length () == 0) {
	    sb.append (s);
	    return;
	}
	sb.append (" ");
	sb.append (s);
    }

    public boolean isPublic () {
	return hasAccessFlag (Opcodes.ACC_PUBLIC);
    }

    public boolean isFinal () {
	return hasAccessFlag (Opcodes.ACC_FINAL);
    }

    public boolean isAbstract () {
	return hasAccessFlag (Opcodes.ACC_ABSTRACT);
    }

    public boolean isInterface () {
	return hasAccessFlag (Opcodes.ACC_INTERFACE);
    }

    public int getNumFields () {
	return numFields;
    }

    public int getNumMethods () {
	return numMethods;
    }

    public int getNumMethodLines () {
	return numMethodLines;
    }

    public int getNumUsedClasses () {
	return numUsedClasses;
    }
}
