package org.khelekore.jsizer;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/** A graph builder that uses full class names to build the graph.
 */
public class HierarchyGraphBuilder implements GraphBuilder {
    public boolean hasFullName () {
	return true;
    }

    public List<TreeNode> buildGraph (List<ClassSize> classSizes,
				      TreeNodeFactory factory) {
	Map<String, Entry> name2Entry = new HashMap<String, Entry> ();
	// Create an Entry for each ClassSize
	for (ClassSize cs : classSizes) {
	    String cn = getName (cs.getClassName ());
	    Entry e = new Entry (cn, cs);
	    name2Entry.put (cn, e);
	}
	// Add child entries
	Set<Entry> roots = new HashSet<Entry> (name2Entry.values ());
	List<Entry> entries = new ArrayList<Entry> (name2Entry.values ());
	for (Entry e : entries) {
	    String sn = e.cs.getSuperName ();
	    if (sn == null) // java.lang.Object
		continue;
	    sn = getName (sn);
	    Entry se = name2Entry.get (sn);
	    if (se == null)  // Non-know super class, add null-entry
		se = addNullEntries (sn, name2Entry, roots);
	    se.addChild (e);
	    roots.remove (e);
	}

	List<Entry> ls = new ArrayList<Entry> (roots);
	Collections.sort (ls, new EntryNameSorter ());

	// convert Entry into TreeNode
	List<TreeNode> ret = new ArrayList<TreeNode> ();
	for (Entry entry : ls) {
	    TreeNode tn = makeTreeNode (factory, entry);
	    ret.add (tn);
	}
	return ret;
    }

    /** Unknown class, add a null-entry and find the superclass until
     *  we either find a known class or we find null (= parent of 
     *  java.lang.Object).
     */
    private Entry addNullEntries (String name,
				  Map<String, Entry> name2Entry,
				  Set<Entry> roots) {
	Entry se = new Entry (name, null);
	name2Entry.put (name, se);
	Entry root = se;
	try {
	    Class<?> c = Class.forName (name);
	    Class<?> s = c.getSuperclass ();
	    while (s != null) {
		name = s.getName ();
		Entry sse = name2Entry.get (name);
		if (sse != null) {
		    sse.addChild (root);
		    root = null;
		    break;
		}
		sse = new Entry (name, null);
		name2Entry.put (name, sse);
		sse.addChild (root);
		root = sse;
		s = s.getSuperclass ();
	    }
	} catch (ClassNotFoundException e) {
	    System.err.println (e);
	}
	if (root != null)
	    roots.add (root);
	return se;
    }

    private static class EntryNameSorter
	implements Comparator<Entry>, Serializable {
	public int compare (Entry e1, Entry e2) {
	    String n1 = e1.name;
	    String n2 = e2.name;
	    return n1.compareTo (n2);
	}
    }

    private String getName (String name) {
	return name.replaceAll ("[$/]", ".");
    }

    private TreeNode getTreeNode (ClassSize cs, TreeNodeFactory factory,
				  String name) {
	if (cs != null)
	    return factory.create (cs, name);
	return factory.create (name);
    }

    private TreeNode makeTreeNode (TreeNodeFactory factory, Entry entry) {
	TreeNode tn = getTreeNode (entry.cs, factory, entry.name);
	if (entry.children != null) {
	    for (Entry e : entry.children) {
		TreeNode cn = makeTreeNode (factory, e);
		tn.addChild (cn);
	    }
	}
	return tn;
    }

    private static class Entry {
	public String name;
	public ClassSize cs;
	public List<Entry> children;

	public Entry (String name, ClassSize cs) {
	    this.name = name;
	    this.cs = cs;
	    this.children = Collections.emptyList ();
	}

	@Override public String toString () {
	    return "Entry{" + name + ", cs: " + cs + "}";
	}

	public void addChild (Entry e) {
	    if (children.isEmpty ())
		children = new ArrayList<Entry> ();
	    children.add (e);
	}
    }
}
