package org.khelekore.jsizer;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import javax.swing.BorderFactory;
import javax.swing.JPanel;
import javax.swing.border.Border;

/** A panel that paints class sizes. */
public class SizePanel extends JPanel {
    private List<ClassSize> classSizes;
    
    private SizeHandler sizeHandler = new SizeHandler ();
    private ColorHandler colorHandler = new ColorHandler (sizeHandler);
    private TreeNodeFactory treeNodeFactory = 
	new DefaultTreeNodeFactory (colorHandler);
    private GraphBuilder graphBuilder = new FullNameGraphBuilder ();
    private List<TreeNode> treeNodes = new ArrayList<TreeNode> ();
    private SizeFilter sizeFilter;
    private ClassNameFilter classNameFilter; 
    private double sx, sy; // current scale factors.
    
    /** never scale graphics up more than this, it looks silly. */
    private static final int MAX_SCALE = 2;
    private static final int HEIGHT_PAD = 20;
    private static final int LEGEND_HEIGHT = 12 * 2;
    private static final int LEGEND_BOX_HEIGHT = LEGEND_HEIGHT / 2;
    private static final int LEGEND_BOX_WIDTH = 45;
    private static final int LEGEND_BOX_SPACING = 5;

    private static Font baseFont = new Font ("Arial", Font.PLAIN, 0);

    private static final long serialVersionUID = 1;
    
    public SizePanel (List<ClassSize> classSizes) {
	this.classSizes = classSizes;
	setPreferredSize (new Dimension (500, 500));
	setOpaque (true);
	setBackground (Color.WHITE);
	Border b1 = BorderFactory.createLineBorder (Color.BLACK, 1);
	Border b2 = BorderFactory.createEmptyBorder (5, 5, 5, 5);
	Border b3 = BorderFactory.createCompoundBorder (b1, b2);
	setBorder (b3);
	setToolTipText ("");
	addMouseListener (new PopupHandler (this));
    }

    public GraphBuilder getGraphBuilder () {
	return graphBuilder;
    }

    public void setGraphBuilder (GraphBuilder gb) {
	graphBuilder = gb;
	buildTree ();
    }

    public TreeNodeFactory getTreeNodeFactory () {
	return treeNodeFactory;
    }

    public void setTreeNodeFactory (TreeNodeFactory factory) {
	this.treeNodeFactory = factory;
	buildTree ();
    }

    public ColorHandler getColorHandler () {
	return colorHandler;
    }

    @Override public String getToolTipText (MouseEvent event) {
	List<TreeNode> tns = getNodesAt (event);
	return getToolTip (tns);
    }

    /** Get the sequence of nodes that form a path down to the mouse position.
     * @return a list of TreeNode or an empty list if no node is at the mouse position.
     */
    public List<TreeNode> getNodesAt (MouseEvent event) {
	Insets insets = getInsets ();
	int mx = event.getX () - insets.left;
	int my = event.getY () - insets.top;
	double x = mx / sx;
	double y = my / sy;
	
	synchronized (this) {
	    for (TreeNode tn : treeNodes)  {
		if (mayContain (tn, x, y)) {
		    List<TreeNode> tns = new ArrayList<TreeNode> ();
		    tns.add (tn);
		    if (buildNodeList (tn, x, y, tns)) 
			return tns;
		}
	    }
	}
	return Collections.emptyList ();
    }

    private String getToolTip (List<TreeNode> tns) {
	if (tns.isEmpty ())
	    return null;
	TreeNode node = tns.get (tns.size () - 1);
	String name = node.getName ();
	if (!graphBuilder.hasFullName ()) {
	    StringBuilder sb = new StringBuilder ();
	    for (TreeNode tn : tns) {
		if (sb.length () > 0)
		    sb.append ('.');
		sb.append (tn.getName ());
	    }
	    name = sb.toString ();
	}
	return getToolTip (node, name);
    }
    
    private boolean mayContain (TreeNode tn, double x, double y) {
	Point2D p = tn.getCenterTop ();
	if (y < p.getY ())
	    return false;
	double w = tn.getWidthWithChildren () / 2.0;
	return ((p.getX () - w) <= x && (p.getX () + w) >= x);
    }

    private boolean buildNodeList (TreeNode p, double x, double y, 
				   List<TreeNode> tns) {
	Rectangle r = p.getRectangle ();
	if (r.contains (x, y))
	    return true;
	for (TreeNode c : p.getChildren ()) {
	    if (mayContain (c, x, y)) {
		tns.add (c);
		if (buildNodeList (c, x, y, tns))
		    return true;
	    }
	}
	return false;
    }

    private String getToolTip (TreeNode tn, String name) {
	if (tn.getZ () < 0) 
	    return "<html><body><b>" + name + "</b></body></html>";
	
	return "<html><body><b>" + name + "</b><br>" +
	    treeNodeFactory.getWidthMetrics () + ": " + tn.getWidth () + 
	    "<br>" + 
	    treeNodeFactory.getHeightMetrics () + ": " + tn.getHeight () + 
	    "<br>" + 
	    treeNodeFactory.getZMetrics () + ": " + tn.getZ () + 
	    "</body></html>";
    }

    public void setSizeFilter (SizeFilter sizeFilter) {
	this.sizeFilter = sizeFilter;
	buildTree ();
    }

    public SizeFilter getSizeFilter () {
	return sizeFilter;
    }

    public void setClassNameFilter (ClassNameFilter classNameFilter) {
	this.classNameFilter = classNameFilter;
	buildTree ();
    }

    public ClassNameFilter getClassNameFilter () {
	return classNameFilter;
    }

    public List<ClassSize> getClassSizes () {
	return Collections.unmodifiableList (classSizes);
    }

    public void rebuild () {
	buildTree ();
    }

    public void setClassSizes (List<ClassSize> classSizes) {
	this.classSizes = classSizes;
	buildTree ();
    }

    private static class ClzNameSorter 
	implements Comparator<ClassSize>, Serializable {
	public int compare (ClassSize cs1, ClassSize cs2) {
	    return cs1.getClassName ().compareTo (cs2.getClassName ());
	}
    }
    
    private synchronized void buildTree () {
	// make sure that we get classes before inner classes... 
	Collections.sort (classSizes, new ClzNameSorter ());
	List<ClassSize> clss = 
	    sizeHandler.findFilteredClassSizes (classSizes, sizeFilter, 
						classNameFilter);

	treeNodes = 
	    graphBuilder.buildGraph (clss, treeNodeFactory);

	// align the top.
	int startHeight = 0;
	for (TreeNode tn : treeNodes) 
	    startHeight = Math.max (startHeight, tn.getGraphicalHeight ());
	sizeHandler.clearSize ();
	for (TreeNode tn : treeNodes) {
	    int startsAt = sizeHandler.getTotalWidth (); 
	    // pad one on each side.
	    int w = tn.getWidthWithChildren () + 2;
	    int centerX = startsAt + w / 2;
	    tn.setPosition (centerX, startHeight);
	    setChildPositions (tn, HEIGHT_PAD);
	    sizeHandler.addToTotalWidth (w);
	}

	repaint ();
    }

    private void setChildPositions (TreeNode parent, int heightPad) {
	Point p = parent.getCenterBottom ();
	int w = p.x - parent.getChildrenWidth () / 2;
	List<TreeNode> ls = parent.getChildren ();
	for (TreeNode tn : ls) {
	    int cw = tn.getWidthWithChildren () + 2;
	    int h = tn.getGraphicalHeight () / 2;
	    tn.setPosition (w + cw / 2, p.y + heightPad + h);
	    int newHeight = p.y + heightPad + tn.getGraphicalHeight ();
	    sizeHandler.maximizeHeight (newHeight);
	    w += cw;
	    setChildPositions (tn, heightPad); 
	}
    }
    
    @Override protected void paintComponent (Graphics g) {
	// clear everything... 
	g.setColor (getBackground ());
	g.fillRect (0, 0, getWidth (), getHeight ());
	synchronized (this) {
	    if (treeNodes.isEmpty ())
		return;
	}

	Insets insets = getInsets ();
	int currentWidth = getWidth () - insets.left - insets.right;
	int currentHeight = 
	    getHeight () - insets.top - insets.bottom - LEGEND_HEIGHT;
	
	Graphics2D g2 = (Graphics2D)g;
	g2.setRenderingHint (RenderingHints.KEY_ANTIALIASING, 
			     RenderingHints.VALUE_ANTIALIAS_ON);

	AffineTransform orig = g2.getTransform ();
	g2.translate (insets.left, insets.top);
	sx = Math.min (MAX_SCALE, 
		       (double)currentWidth / sizeHandler.getTotalWidth ());
	sy = Math.min (MAX_SCALE, 
		       (double)currentHeight / sizeHandler.getTotalHeight ());
	g2.scale (sx, sy);
	synchronized (this) {
	    for (TreeNode tn : treeNodes)
		drawNodes (g2, tn);
	    AffineTransform paintTransform = g2.getTransform ();
	    g2.setTransform (orig);
	    drawPackageNames (g2, insets, paintTransform);
	    paintLegend (g2, insets);
	}
    }

    private void paintLegend (Graphics2D g2, Insets insets) {
	// kind of silly to draw lots of black boxes..
	if (sizeHandler.getMaxLines () == 0) 
	    return;
	g2.translate (insets.left, insets.top);
	Font f = g2.getFont ().deriveFont (Font.BOLD, LEGEND_BOX_HEIGHT);
	g2.setFont (f);
	int y = getHeight () - insets.top - insets.bottom - LEGEND_BOX_HEIGHT;
	int numBoxes = findNumBoxes (insets);
	if (numBoxes > 0)
	    for (int i = 0; i < numBoxes; i++)
		drawColorBox (g2, i, numBoxes, y); 
	g2.setColor (Color.BLACK);
	
	y--; // create extra space...
	String s = "Max fields: " + sizeHandler.getMaxFields ();
	Rectangle2D r = 
	    g2.getFont ().getStringBounds (s, g2.getFontRenderContext ());
	g2.drawString (s, 0, y); 
	int x = (int)r.getWidth () + LEGEND_BOX_SPACING;
	String t = "Max methods: " + sizeHandler.getMaxMethods ();
	r = g2.getFont ().getStringBounds (t, g2.getFontRenderContext ());
	g2.drawString (t, x, y); 
	String u = "Max used classes: " + sizeHandler.getMaxUsedClasses ();
	x = x + (int)r.getWidth () + LEGEND_BOX_SPACING;
	g2.drawString (u, x, y);
	g2.translate (-insets.left, -insets.top);
    }
    
    // use component width to figure out number of boxes, 
    // never more than 10 though.
    private int findNumBoxes (Insets insets) {
	int numBoxes = // last space can be ignored...
	    (getWidth () - insets.left - insets.right + LEGEND_BOX_SPACING) / 
	    (LEGEND_BOX_WIDTH + LEGEND_BOX_SPACING);
	return Math.min (numBoxes, 10);
    }
	
    private void drawColorBox (Graphics2D g2, int i, int max, int y) {
	int maxLines = sizeHandler.getMaxLines ();
	int lines = max > 1 ? maxLines * i / (max - 1) : maxLines;
	Color c = colorHandler.getColor (lines);
	g2.setColor (c);
	int x = (LEGEND_BOX_WIDTH + LEGEND_BOX_SPACING) * i;
	g2.fillRect (x, y, LEGEND_BOX_WIDTH, LEGEND_BOX_HEIGHT);
	c = getTextColor (c);
	g2.setColor (c);
	String s = Integer.toString (lines);
	Rectangle2D r = 
	    g2.getFont ().getStringBounds (s, g2.getFontRenderContext ());
	int sx = x + LEGEND_BOX_WIDTH / 2 - (int)(r.getWidth () / 2);
	g2.drawString (s, sx, y + LEGEND_BOX_HEIGHT - 1);
    }

    private Color getTextColor (Color c) {
	float[] hsbVals = 
	    Color.RGBtoHSB (c.getRed (), c.getGreen (), c.getBlue (), null);
	return hsbVals[2] > 0.5 ? Color.BLACK : Color.WHITE;
    }

    private void drawNodes (Graphics2D g2, TreeNode tn) {
	Rectangle r = tn.getRectangle ();
	g2.setColor (tn.getColor ());
	g2.fillRect (r.x, r.y, r.width, r.height);
	synchronized (this) {
	    Point pb = tn.getCenterBottom ();
	    for (TreeNode ctn : tn.getChildren ()) {
		g2.setColor (Color.YELLOW);
		Point pt = ctn.getCenterTop ();
		g2.drawLine (pb.x, pb.y, pt.x, pt.y);
		drawNodes (g2, ctn);
	    }
	}
    }

    private void drawPackageNames (Graphics2D g2, Insets insets, 
				   AffineTransform paintTransform) {
	double size = 10 * sy;
	if (size > 4) {
	    Font f = baseFont.deriveFont ((float)size);
	    g2.setFont (f);
	    g2.setColor (Color.BLACK);
	    for (TreeNode tn : treeNodes)
		drawPackageName (g2, tn, insets);
	}
	for (TreeNode tn : treeNodes)
	    drawClassName (g2, tn, insets);
    }
    
    private void drawPackageName (Graphics2D g2, TreeNode tn, Insets insets) {
	if (tn.isPackage ()) {
	    String name = getShortName (tn);
	    FontMetrics fm = g2.getFontMetrics ();
	    double width = fm.stringWidth (name) / sx;
	    if (width < tn.getWidthWithChildren () / 2.0) {
		Point2D p = tn.getCenterBottom ();
		double x;
		if (p.getX () * sx > getWidth () / 2.0) {
		    x = p.getX () - tn.getGraphicalWidth () / 2.0;
		    x = x * sx + insets.left - 1 - width * sx ;
		} else {
		    x = p.getX () + tn.getGraphicalWidth () / 2.0;
		    x = x * sx  + insets.left + 1;
		}
		double y = p.getY ();
		y = y * sy + insets.top - 1;
		g2.drawString (name, (float)x, (float)y);
	    }
	}
	for (TreeNode ctn : tn.getChildren ()) 
	    drawPackageName (g2, ctn, insets);
    }

    private String getShortName (TreeNode tn) {
	String name = tn.getName ();
	if (graphBuilder.hasFullName ()) {
	    int i = name.lastIndexOf ('.');
	    if (i > -1)
		name = name.substring (i + 1);
	}
	return name;
    }

    private void drawClassName (Graphics2D g2, TreeNode tn, Insets insets) {
	if (!tn.isPackage ()) {
	    double gw = tn.getGraphicalWidth () * sx;
	    double size = Math.min (13, gw);
	    while (size > 8) {
		double gh = tn.getGraphicalHeight () * sy;
		String name = getShortName (tn);
		Font f = baseFont.deriveFont ((float)size);
		size -= 1;
		g2.setFont (f);
		FontMetrics fm = g2.getFontMetrics ();
		double width = fm.stringWidth (name) + 2;
		if (width < gh) {
		    Point2D p = tn.getCenterTop ();
		    AffineTransform orig = g2.getTransform ();
		    g2.translate (insets.left, insets.top);
		    AffineTransform at = new AffineTransform (g2.getTransform ());
		    at.translate (p.getX () * sx, p.getY () * sy);
		    at.quadrantRotate (1);
		    at.translate (1, fm.getAscent () / 2 - 1);
		    g2.setTransform (at);
		    Color c = 
			getTextColor (colorHandler.getColor (tn.getZ ()));
		    g2.setColor (c);
		    g2.drawString (name, 0, 0);
		    g2.setTransform (orig);
		    break;
		}
	    }
	}
	
	for (TreeNode ctn : tn.getChildren ()) 
	    drawClassName (g2, ctn, insets);
    }

    public void addColor (String classname, Color color) {
	colorHandler.addColor (classname, color);
    }

    public void removeColor (String classname) {
	colorHandler.removeColor (classname);
    }

    public void clearColors () {
	colorHandler.clearColors ();
    }    
}
